from __future__ import annotations

import inspect

import numpy as np
import pandas as pd


def lissajous(
    time,
    radius=1.0,
    speed=None,
    width=None,
    height=None,
    freq_ratio=1.193,
):  # noqa
    width = width or radius / 2
    height = height or width
    speed = speed or width / 10

    freq = speed / np.sqrt((width * freq_ratio) ** 2 + height**2)

    x = width * np.cos(freq_ratio * freq * time)
    y = height * np.sin(freq * time)

    return np.stack([x, y])


def double_circle(time, speed=None, radius=1.0, ratio=0.5, freq_ratio=1.7):
    speed = speed or radius / 10

    a = radius / (1 + 1 / ratio)
    b = a / ratio

    phase = time * speed / np.maximum(a + b * freq_ratio, 1e-16)  # do not divide by zero!

    x_p = a * np.sin(phase) + b * np.sin(phase * freq_ratio)
    y_p = a * np.cos(phase) + b * np.cos(phase * freq_ratio)

    return np.stack([x_p, y_p])


def daisy_from_phase(phase, a, b, petals, miss_freq):
    x_p = a * np.cos(petals * phase) * np.sin(phase) + b * np.sin(petals * phase) * np.cos(miss_freq * phase)
    y_p = a * np.cos(petals * phase) * np.cos(phase) + b * np.sin(petals * phase) * np.sin(miss_freq * phase)
    X = np.stack([x_p, y_p])
    return (a + b) * X / np.sqrt(np.square(X).sum(axis=0).max())


def daisy(
    time,
    radius=1.0,
    speed=None,
    petals=7 / np.pi,
    miss_factor=0.15,
    miss_freq=np.sqrt(2),
):  # noqa
    if radius > 0:
        speed = speed or radius / 10
        a = radius / (1 + miss_factor)
        b = a * miss_factor
        max_speed = 0.0
        dp_dt = (speed / radius) if radius > 0.0 else 0.0

        for _ in range(4):
            phase = dp_dt * time
            test_x, test_y = daisy_from_phase(phase, a=a, b=b, petals=petals, miss_freq=miss_freq)
            vx2 = (np.gradient(test_x) / np.gradient(time)) ** 2
            vy2 = (np.gradient(test_y) / np.gradient(time)) ** 2
            max_speed = np.sqrt(vx2 + vy2).max()

            if np.abs(np.log(max_speed / speed)) > 0.01:
                dp_dt *= speed / max_speed
            else:
                break

        return daisy_from_phase(phase, a=a, b=b, petals=petals, miss_freq=miss_freq)

    return np.zeros((2, len(time)))


def smooth_sawtooth(p, throw=1, delta=0.01):
    return throw * (1 - 2 * np.arccos((1 - delta) * np.sin(p)) / np.pi)


def back_and_forth(t, radius=1, x_throw=None, y_throw=0, speed=1.0, max_accel=np.inf, d=0.01):
    x_throw = x_throw or radius

    factor = 1 / (1 - 2 * np.arccos(1 - d) / np.pi)

    throw = factor * np.sqrt(x_throw**2 + y_throw**2)

    a = np.pi * speed / (2 * throw * (1 - d))
    b = np.sqrt(np.pi * max_accel * np.sqrt(2 * d - d**2) / (2 * throw * (1 - d)))

    dp_dt = np.minimum(a, b)

    x = factor * smooth_sawtooth(dp_dt * t, throw=x_throw, delta=d)
    y = factor * smooth_sawtooth(dp_dt * t, throw=y_throw, delta=d)

    return np.stack([x, y])


# def grid(time, radius=1, speed=None, n=17, turnaround_time=5):  # noqa
#     speed = speed or radius / 5

#     duration = np.ptp(time)

#     xs = []
#     ys = []

#     minor_axis = "x"

#     timestep = (2 * radius / speed) / (n - 1)

#     side = np.linspace(-radius, radius, n)

#     while timestep * (len(xs) - n) <= duration:
#         minor = []
#         major = []

#         for i, y in enumerate(side):
#             minor.extend(side[:: (-1) ** (i + 1)])
#             major.extend(np.repeat(y, n))

#         minor.pop(-1), major.pop(-1)

#         if minor_axis == "x":
#             xs.extend(minor)
#             ys.extend(major)
#         else:
#             xs.extend(major)
#             ys.extend(minor)

#         minor_axis = "y" if minor_axis == "x" else "x"

#     offsets = sp.interpolate.interp1d(
#         timestep * np.arange(len(xs)) + time.min(),
#         np.c_[xs, ys].T,
#         bounds_error=False,
#         fill_value="extrapolate",
#     )(time)

#     return sp.ndimage.gaussian_filter1d(
#         offsets,
#         sigma=turnaround_time / timestep,
#         axis=-1,
#     )


# def smooth_sawtooth(phase, width=0.5, smoothness=0.5):
#     smooth_phase = sp.ndimage.gaussian_filter1d(
#         phase,
#         sigma=smoothness * np.gradient(phase).mean(),
#     )
#     smooth_sawtooth = sp.signal.sawtooth(smooth_phase, width=width)
#     return 2 * (smooth_sawtooth - smooth_sawtooth.min()) / smooth_sawtooth.ptp() - 1


# def raster(time, radius=1, height=None, speed=0.5, n=16, turnaround_time=0.5):
#     width = 2 * radius
#     height = height if height is not None else width

#     sample_rate = 1 / np.gradient(time).mean()

#     start_time = time.min()

#     ts, xs, ys = [], [], []

#     n_scans = 2 * n + 1
#     phase = np.linspace(0, np.pi * n_scans, n_scans * 256)
#     raster_period = 2 * n_scans * np.sqrt(width**2 + (height / n_scans) ** 2) / speed

#     while start_time < time.max():
#         xs.extend(0.5 * width * sp.signal.sawtooth(phase, width=0.5))
#         ys.extend(height * np.linspace(0.5, -0.5, len(phase)))
#         ts.extend(start_time + np.linspace(0, raster_period, len(phase)))
#         start_time = ts[-1] + np.sqrt(width**2 + height**2) / speed

#     offsets = sp.interpolate.interp1d(ts, np.c_[xs, ys].T)(time)

#     return sp.ndimage.gaussian_filter1d(
#         offsets,
#         sigma=turnaround_time * sample_rate,
#         axis=-1,
#     )


# def back_and_forth(time, speed=1, radius=5, turnaround_time=0.5):  # noqa
#     return raster(
#         time,
#         speed=speed,
#         radius=radius,
#         height=0,
#         turnaround_time=turnaround_time,
#     )


def stare(time):
    return np.zeros((2, *time.shape))


# def get_constant_speed_offsets(
#     pattern,
#     duration,
#     sample_rate,
#     speed,
#     eps=1e-6,
#     **scan_options,
# ):
#     """This should be cythonized maybe."""

#     speed_per_sample = speed / sample_rate

#     def phase_coroutine():
#         p = 0.0

#         for _ in range(int(duration * sample_rate)):
#             x0, y0 = pattern(p - 0.5 * eps, **scan_options)
#             x1, y1 = pattern(p + 0.5 * eps, **scan_options)

#             ds_dp = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2) / eps  # speed per phase

#             dp = speed_per_sample / ds_dp
#             p += dp

#             yield p

#     return pattern(np.array([p for p in phase_coroutine()]), **scan_options)


scan_patterns = {
    "stare": {"aliases": [], "generator": stare},
    "daisy": {"aliases": ["daisy_scan"], "generator": daisy},
    "lissajous": {"aliases": ["lissajous_box"], "generator": lissajous},
    # "raster": {"aliases": [], "generator": raster},
    "back_and_forth": {"aliases": ["back-and-forth"], "generator": back_and_forth},
    # "grid": {"aliases": [], "generator": grid},
    "double_circle": {"aliases": [], "generator": double_circle},
}

for key in scan_patterns:
    scan_patterns[key]["signature"] = str(inspect.signature(scan_patterns[key]["generator"]))

scan_patterns = pd.DataFrame(scan_patterns).T


def get_scan_pattern_generator(name):
    for index, entry in scan_patterns.iterrows():
        if (name == index) or (name in entry.aliases):
            return entry.generator

    raise ValueError(f"Invalid pattern name '{name}'.")



from __future__ import annotations

import copy
import gc
import json
import logging
import time as ttime

import arrow
import h5py
import numpy as np
import pandas as pd
import scipy as sp
from astropy.io import fits
from astropy.time import Time
from dask import array as da

from ..array import Array, ArrayList
from ..band import get_band
from ..coords import Coordinates
from ..instrument import get_instrument
from ..io import DEFAULT_TIME_FORMAT, humanize_time
from ..plotting import plot_tod, twinkle_plot
from ..site import get_site
from ..spectrum import AtmosphericSpectrum

logger = logging.getLogger("maria")


class TOD:
    """
    Time-ordered data. This has per-detector pointing and data.
    """

    def __init__(
        self,
        data: dict,
        weight: float = None,
        coords: Coordinates = None,
        units: str = "K_RJ",
        dets: ArrayList = None,
        distributed: bool = True,
        dtype: type = np.float32,
        metadata: dict = {},
    ):
        self.coords = coords
        self.dets = dets
        self.header = fits.header.Header()
        self.units = units
        self.dtype = dtype
        self.metadata = metadata

        self.data = {}

        for field, field_data in data.items():
            if field_data.ndim != 2:
                raise ValueError("Only two-dimensional TODs are currently supported.")

            self.data[field] = da.asarray(field_data, dtype=dtype) if distributed else field_data.astype(dtype)

        # sort them alphabetically
        self.data = {k: self.data[k] for k in sorted(list(self.fields))}

        if weight is None:
            weight = np.ones(self.signal.shape)
        if distributed:
            weight = da.asarray(weight)

        self.weight = weight

    @property
    def spectrum(self):
        if not hasattr(self, "_spectrum"):
            if "region" in self.metadata:
                self._spectrum = AtmosphericSpectrum(self.metadata["region"])
            else:
                self._spectrum = None
        return self._spectrum

    @property
    def boresight(self):
        if not hasattr(self, "_boresight"):
            self._boresight = self.coords.boresight()
        return self._boresight

    def calibration_kwargs(self, band=None):
        band_mask = self.dets.band_name == band.name
        kwargs = {
            "elevation": self.el[band_mask] if band else self.el,
            "polarized": ~np.isnan(self.dets.pol_angle[band_mask]).all(),
        }
        if self.metadata["atmosphere"]:
            kwargs["spectrum"] = self.spectrum
            kwargs["zenith_pwv"] = self.metadata["pwv"]
            kwargs["base_temperature"] = self.metadata["base_temperature"]

        else:
            kwargs["spectrum"] = None

        return kwargs

    def to(self, units: str):
        """
        Convert to a different set of units.
        """

        cal_start_s = ttime.monotonic()

        # make sure that all Array have a band that the TOD knows about
        for band_name in np.unique(self.dets.band_name):
            if band_name not in self.dets.bands.name:
                raise ValueError(f"No band defined for detector with band '{band_name}'.")

        content = self.content
        for band in self.dets.bands:
            band_mask = self.dets.band_name == band.name

            if band_mask.sum() == 0:
                continue

            # this is to handle transmission
            cal = band.cal(f"{self.units} -> {units}", **self.calibration_kwargs(band))

            for field in self.fields:
                content["data"][field][band_mask] = cal(self.data[field][band_mask])

        content["units"] = units

        logger.debug(f'Converted {self} to units "{units}" in {humanize_time(ttime.monotonic() - cal_start_s)}.')

        return TOD(**content)

    @property
    def shape(self):
        return self.signal.shape

    @property
    def fields(self) -> list:
        return sorted(list(self.data.keys()))

    @property
    def signal(self) -> da.Array:
        return sum([self.data[field] for field in self.fields])

    @property
    def duration(self) -> float:
        return np.ptp(self.time)

    @property
    def dt(self) -> float:
        return float(np.gradient(self.time, axis=-1).mean())

    @property
    def sample_rate(self) -> float:
        return float(1 / self.dt)

    @property
    def fs(self) -> float:
        return self.sample_rate

    @property
    def nd(self) -> int:
        return int(self.signal.shape[0])

    @property
    def nt(self) -> int:
        return int(self.signal.shape[-1])

    @property
    def start(self):
        return arrow.get(self.time.max())

    @property
    def end(self):
        return arrow.get(self.time.max())

    def subset(
        self,
        det_mask: bool | int = None,
        time_mask: bool | int = None,
        band: str = None,
        fields: list = None,
    ):
        det_mask = det_mask or np.arange(self.nd)
        time_mask = time_mask or np.arange(self.nt)
        fields = fields or self.fields

        if band is not None:
            det_mask = self.dets.band_name == band
            # if not det_mask.sum() > 0:
            #     raise ValueError(f"There are no Array for band '{band}'.")

        if time_mask is not None:
            if len(time_mask) != self.nt:
                raise ValueError("The detector mask must have shape (n_dets,).")

        if det_mask is not None:
            if not (len(det_mask) == self.nd):
                raise ValueError("The detector mask must have shape (n_dets,).")

        content = self.content
        content.update(
            {
                "data": {field: self.data[field][det_mask][..., time_mask].compute() for field in fields},
                "weight": self.weight[det_mask][..., time_mask].compute(),
                "coords": self.coords[det_mask][..., time_mask],
                "dets": self.dets._subset(det_mask),
            }
        )

        return TOD(**content)

    @property
    def time(self):
        return self.coords.t

    @property
    def earth_location(self):
        return self.coords.earth_location

    @property
    def lat(self):
        return np.round(self.earth_location.lat.deg, 6)

    @property
    def lon(self):
        return np.round(self.earth_location.lon.deg, 6)

    @property
    def alt(self):
        return np.round(self.earth_location.height.value, 6)

    @property
    def az(self):
        return self.coords.az

    @property
    def el(self):
        return self.coords.el

    @property
    def ra(self):
        return self.coords.ra

    @property
    def dec(self):
        return self.coords.dec

    @property
    def azim_phase(self):
        return np.pi * (sp.signal.sawtooth(2 * np.pi * self.time / self.azim_scan_period, width=1) + 1)

    @property
    def turnarounds(self):
        azim_grad = sp.ndimage.gaussian_filter(np.gradient(self.azim), sigma=16)
        return np.where(np.sign(azim_grad[:-1]) != np.sign(azim_grad[1:]))[0]

    def splits(self, target_split_time: float = None):
        if target_split_time is None:
            return list(zip(self.turnarounds[:-1], self.turnarounds[1:]))
        else:
            fs = self.fs
            splits_list = []
            for s, e in self.splits(target_split_time=None):
                split_time = self.time[e] - self.time[s]  # total time in the split
                n_splits = int(
                    np.ceil(split_time / target_split_time),
                )  # number of new splits
                n_split_samples = int(
                    target_split_time * fs,
                )  # number of samples per new split
                for split_start in np.linspace(s, e - n_split_samples, n_splits).astype(
                    int,
                ):
                    splits_list.append(
                        (split_start, np.minimum(split_start + n_split_samples, e)),
                    )
            return splits_list

    def to_fits(self, fname, format="MUSTANG-2", overwrite=True):
        """
        Save the TOD to a fits file
        """

        if format.lower() == "mustang-2":
            header = fits.header.Header()

            header["AZIM"] = (self.coords.center("az/el")[0], "radians")
            header["ELEV"] = (self.coords.center("az/el")[1], "radians")
            header["BMAJ"] = (9.0, "arcsec")
            header["BMIN"] = (9.0, "arcsec")
            header["BPA"] = (0.0, "degrees")
            header["NDETS"] = self.dets.n
            header["JDSTART"] = Time(self.time[0], format="unix").jd

            header["SITELAT"] = (self.lat, "Site Latitude")
            header["SITELONG"] = (self.lon, "Site Longitude")
            header["SITEELEV"] = (self.alt, "Site elevation (meters)")

            col01 = fits.Column(
                name="DX   ",
                format="E",
                array=self.coords.ra.flatten(),
                unit="radians",
            )
            col02 = fits.Column(
                name="DY   ",
                format="E",
                array=self.coords.dec.flatten(),
                unit="radians",
            )

            tod_rj = self.to(units="K_RJ")
            col03 = fits.Column(
                name="FNU  ",
                format="E",
                array=tod_rj.signal.compute().flatten(),
                unit=tod_rj.units,
            )
            col04 = fits.Column(name="UFNU ", format="E")
            col05 = fits.Column(
                name="TIME ",
                format="E",
                array=((self.time - self.time[0]) * np.ones_like(self.coords.ra)).flatten(),
                unit="s",
            )
            col06 = fits.Column(name="COL  ", format="I")
            col07 = fits.Column(name="ROW  ", format="I")
            col08 = fits.Column(
                name="PIXID",
                format="I",
                array=(
                    np.arange(len(self.coords.ra), dtype=np.int16).reshape(-1, 1) * np.ones_like(self.coords.ra)
                ).flatten(),
            )
            col09 = fits.Column(
                name="SCAN ",
                format="I",
                array=np.zeros_like(self.coords.ra).flatten(),
            )
            col10 = fits.Column(name="ELEV ", format="E")

            hdu = fits.BinTableHDU.from_columns(
                [col01, col02, col03, col04, col05, col06, col07, col08, col09, col10],
                header=header,
            )

            hdu.writeto(fname, overwrite=overwrite)

    def to_hdf(self, fname):
        with h5py.File(fname, "w") as f:
            f.createdataset(fname)

    @staticmethod
    def from_fits(fname: str, format: str, **kwargs):
        if format.lower() == "mustang-2":
            return TOD._from_mustang2(fname=fname, **kwargs)

        if format.lower() == "atlast":  # hopefull
            ...

        if format.lower() == "act":
            ...

    @classmethod
    def _from_mustang2(cls, fname: str, index: int = 1):
        with fits.open(fname) as f:
            raw = f[index].data

            det_uids, det_counts = np.unique(raw["PIXID"], return_counts=True)

            if det_counts.std() > 0:
                raise ValueError("Cannot reshape a ragged TOD.")

            n_dets = len(det_uids)
            n_samp = det_counts.max()

            data = {"data": raw["FNU"].astype("float32").reshape((n_dets, n_samp))}

            ra = raw["dx"].astype(float).reshape((n_dets, n_samp))  # rad
            dec = raw["dy"].astype(float).reshape((n_dets, n_samp))

            t = raw["time"].astype(float).reshape((n_dets, n_samp)).mean(axis=0)

            if "JDSTART" in f[index].header:
                t += Time(f[index].header["JDSTART"], format="jd").unix

            site = get_site("green_bank")

            boresight = Coordinates(
                t=t,
                phi=ra,
                theta=dec,
                earth_location=site.earth_location,
                frame="ra/dec",
            )

            # building array class
            dets_dict = {
                "sky_x": ra[:, 0] - ra[:, 0].mean(),  # in ra/dec frame
                "sky_y": dec[:, 0] - dec[:, 0].mean(),  # in ra/dec frame
                "band_name": len(dec[:, 0]) * ["m2/f093"],
            }

            dets_df = pd.DataFrame(dets_dict)
            _array = get_instrument("mustang-2").arrays[0]

            for col in _array.dets.columns:
                if col in dets_df.columns:
                    continue
                dets_df[col] = _array.dets.iloc[0][col]

            a = Array(name="mustang2", dets=dets_df, bands=[get_band("m2/f093")])

            metadata = {
                "atmosphere": False,
                "altitude": float(site.altitude.m),
                "region": site.region,
                "Real_obs": True,
                "base_temperature": f[index].header.get("TAMBIENT", None),
            }

            tod = TOD(
                data=data,
                dets=a,
                coords=boresight,
                units="K_RJ",
                metadata=metadata,
            )

            gc.collect()
            return tod

    def plot(self, detrend=True, max_dets: int = 1, n_freq_bins: int = 1024):
        plot_tod(
            self,
            detrend=detrend,
            n_freq_bins=n_freq_bins,
            max_dets=max_dets,
        )

    def twinkle(self, filename=None, **kwargs):
        twinkle_plot(
            self,
            filename=filename,
            **kwargs,
        )

    def __getattr__(self, attr):
        if attr in self.fields:
            return self.data[attr]
        raise AttributeError(f"'TOD' object has no attribute '{attr}'")

    def __repr__(self):
        parts = []
        parts.append(f"shape={self.shape}")
        parts.append(f"fields={repr(self.fields)}")
        parts.append(f"units={repr(self.units)}")
        parts.append(f"start={self.start.format(DEFAULT_TIME_FORMAT)}")
        parts.append(f"duration={self.duration:.01f}s")
        parts.append(f"sample_rate={self.sample_rate:.01f}Hz")
        parts.append(f"metadata={self.metadata}")
        return f"TOD({', '.join(parts)})"

    @property
    def content(self):
        res = {"data": {}}
        for field in self.fields:
            res["data"][field] = copy.deepcopy(self.data[field])
        for key in ["coords", "weight", "units", "dets", "dtype", "metadata"]:
            if hasattr(self, key):
                res[key] = getattr(self, key)
        return res

    def copy(self):
        """
        Copy yourself.
        """
        return TOD(**self.content)


def check_nested_keys(keys_found, data, keys):
    for key in data.keys():
        for i in range(len(keys)):
            if keys[i] in data[key].keys():
                keys_found[i] = True


def check_json_file_for_key(keys_found, file_path, *keys_to_check):
    with open(file_path) as json_file:
        data = json.load(json_file)
        return check_nested_keys(keys_found, data, keys_to_check)


def test_multiple_json_files(files_to_test, *keys_to_find):
    keys_found = np.zeros(len(keys_to_find)).astype(bool)

    for file_path in files_to_test:
        check_json_file_for_key(keys_found, file_path, *keys_to_find)

    if np.sum(keys_found) != len(keys_found):
        raise KeyError(np.array(keys_to_find)[~keys_found])



from __future__ import annotations

import logging
import os
import time as ttime

import dask.array as da
import numpy as np
import scipy as sp
from jax import scipy as jsp
from tqdm import tqdm

from ..constants import k_B
from ..io import DEFAULT_BAR_FORMAT, humanize_time

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")


class AtmosphereMixin:
    def _simulate_atmosphere(self, obs):
        # this produces self.atmosphere.zenith_scaled_pwv at lower res
        # which we use to compute emission and opacity
        obs.atmosphere.simulate_pwv(instrument=obs.instrument)

        # upsample to the sim resolution
        obs.zenith_scaled_pwv = da.from_array(
            sp.interpolate.interp1d(
                obs.atmosphere.coords.t,
                obs.atmosphere.zenith_scaled_pwv,
                bounds_error=False,
                fill_value="extrapolate",
            )(obs.coords.t),
        )

    def _compute_atmospheric_loading(self, obs):
        atmosphere_loading = np.zeros(obs.atmosphere.zenith_scaled_pwv.shape, dtype=self.dtype)

        bands_pbar = tqdm(
            obs.instrument.dets.bands,
            desc="Computing atmospheric emission",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
            ncols=250,
        )
        for band in bands_pbar:
            emission_s = ttime.monotonic()

            bands_pbar.set_postfix({"band": band.name})

            band_index = obs.instrument.dets.mask(band_name=band.name)

            # the 1e12 is for picowatts
            det_power_grid = (
                1e12
                * k_B
                * np.trapezoid(
                    obs.atmosphere.spectrum._emission * band.passband(obs.atmosphere.spectrum.side_nu),
                    obs.atmosphere.spectrum.side_nu,
                    axis=-1,
                )
            )

            band_power_interpolator = jsp.interpolate.RegularGridInterpolator(
                obs.atmosphere.spectrum.points[:3],
                det_power_grid,
            )

            atmosphere_loading[band_index] = band_power_interpolator(
                (
                    obs.atmosphere.weather.temperature[0],
                    obs.atmosphere.zenith_scaled_pwv[band_index],
                    obs.atmosphere.coords.el[band_index].clip(max=np.pi / 2),
                ),
            )

            logger.debug(
                f"Computed atmospheric emission for band {band.name} in {humanize_time(ttime.monotonic() - emission_s)}."
            )

        # upsample to the sim resolution
        return da.asarray(
            sp.interpolate.interp1d(
                obs.atmosphere.coords.t,
                atmosphere_loading,
                kind="quadratic",
                bounds_error=False,
                fill_value="extrapolate",
            )(obs.coords.t),
            dtype=self.dtype,
        )



from __future__ import annotations

import numpy as np
import scipy as sp

# def extrude(
#     values: np.array,
#     A: np.array,
#     B: np.array,
#     n_steps: int,
#     n_i: int,
#     n_j: int,
#     i_sample_index: int,
#     j_sample_index: int,
# ):
#     # muy rapido
#     BUFFER = np.zeros((n_steps + n_i) * n_j)
#     BUFFER[n_steps * n_j :] = values

#     # remember that (e, c) -> n_c * e + c
#     for buffer_index in np.arange(n_steps)[::-1]:
#         BUFFER[buffer_index * n_j + np.arange(n_j)] = A @ BUFFER[
#             n_j * (buffer_index + 1 + i_sample_index) + j_sample_index
#         ] + B @ np.random.standard_normal(size=n_j)

#     return BUFFER[: n_steps * n_j]


# def get_rotation_matrix_2d(a):
#     S = np.atleast_1d(a)[..., None, None] * np.array([[0.0, 1.0], [-1.0, 0.0]])
#     R = sp.linalg.expm(S)
#     return R.reshape(*np.shape(a), 2, 2)


# def get_rotation_matrix_3d(angles, axis=0):
#     shaped_angles = np.atleast_1d(angles)
#     R = np.ones(shaped_angles.shape)[..., None, None] * np.eye(3)
#     j = [i for i in [0, 1, 2] if not i == axis]
#     j0, j1 = np.meshgrid(j, j)
#     R[..., j0, j1] = sp.linalg.expm(
#         np.array([[0, -1], [+1, 0]]) * shaped_angles[..., None, None],
#     )
#     return R.reshape(*np.shape(angles), 3, 3)


def fast_psd_inverse(M):
    """
    About twice as fast as np.linalg.inv for large, PSD matrices.
    """

    cholesky, dpotrf_info = sp.linalg.lapack.dpotrf(M)
    invM, dpotri_info = sp.linalg.lapack.dpotri(cholesky)
    return np.where(invM, invM, invM.T)



from __future__ import annotations

import logging
import os
import time as ttime
import warnings
from collections.abc import Sequence
from typing import Mapping

import numpy as np
import scipy as sp

from ..coords import FRAMES, Frame, get_center_phi_theta
from ..instrument import BandList
from ..io import humanize_time
from ..map import ProjectedMap
from ..tod import TOD
from ..units import Quantity

# np.seterr(invalid="ignore")

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class BaseProjectionMapper:
    """
    The base class for mapping.
    """

    def __init__(
        self,
        center: tuple[float, float],
        stokes: str,
        width: float,
        height: float,
        resolution: float,
        frame: str,
        units: str,
        degrees: bool,
        calibrate: bool,
        tods: Sequence[TOD],
        tod_preprocessing: Mapping,
        map_postprocessing: Mapping,
    ):
        center = center or np.degrees(get_center_phi_theta(*np.stack([tod.coords.center(frame="ra/dec") for tod in tods]).T))
        height = height or width

        self.resolution = Quantity(resolution, "deg" if degrees else "rad")
        self.center = Quantity(center, "deg" if degrees else "rad")
        self.width = Quantity(width, "deg" if degrees else "rad")
        self.height = Quantity(height, "deg" if degrees else "rad")
        self.degrees = degrees
        self.calibrate = calibrate
        self.frame = Frame(frame)
        self.units = units
        self.stokes = stokes
        self.tod_preprocessing = tod_preprocessing
        self.map_postprocessing = map_postprocessing

        self.bands = BandList([])

        self.tods = []
        self.add_tods(tods)

        self.map = ProjectedMap(
            data=np.zeros((self.n_y, self.n_x)),
            weight=np.ones((self.n_y, self.n_x)),
            stokes=self.stokes,
            nu=self.bands.center,
            resolution=self.resolution,
            center=self.center,
            degrees=False,
            frame=self.frame.name,
            units=self.units,
        )

    def add_tods(self, tods):
        for tod in np.atleast_1d(tods):
            self.tods.append(tod.process(config=self.tod_preprocessing).to(self.units))

            for band in tod.dets.bands:
                self.bands.add(band)

        self.initialize_mapper()

    def initialize_mapper(self):
        raise NotImplementedError()

    @property
    def map_shape(self):
        return (len(self.stokes), self.n_x, self.n_y)

    @property
    def n_x(self):
        return int(np.maximum(1, self.width / self.resolution))

    @property
    def n_y(self):
        return int(np.maximum(1, self.height / self.resolution))

    @property
    def x_bins(self):
        return np.linspace(-0.5 * self.width.rad, 0.5 * self.width.rad, self.n_x + 1)

    @property
    def y_bins(self):
        return np.linspace(0.5 * self.height.rad, -0.5 * self.height.rad, self.n_y + 1)

    def plot(self):
        if not hasattr(self, "map"):
            raise RuntimeError("Mapper has not been run yet.")
        self.map.plot()

    @property
    def n_stokes(self):
        return len(self.stokes)

    @property
    def n_bands(self):
        return len(self.bands)

    def _run(self):
        raise ValueError("Not implemented!")

    def run(self):
        if not len(self.tods):
            raise RuntimeError("This mapper has no TODs!")

        map_data = self._run()

        if "gaussian_filter" in self.map_postprocessing.keys():
            sigma = self.map_postprocessing["gaussian_filter"]["sigma"]

            map_data["sum"] = sp.ndimage.gaussian_filter(map_data["sum"], sigma=(0, 0, sigma, sigma))
            map_data["wgt"] = sp.ndimage.gaussian_filter(map_data["wgt"], sigma=(0, 0, sigma, sigma))

        if "median_filter" in self.map_postprocessing.keys():
            size = self.map_postprocessing["median_filter"]["size"]
            map_data["sum"] = sp.ndimage.median_filter(
                map_data["sum"],
                size=size,
                axes=(-2, -1),
            )
            map_data["wgt"] = sp.ndimage.median_filter(
                map_data["wgt"],
                size=size,
                axes=(-2, -1),
            )

        for stokes_index, stokes in enumerate(self.stokes):
            for nu_index, nu in enumerate(self.bands.center):
                if map_data["wgt"][stokes_index, nu_index].sum() == 0:
                    logger.warning(f"No counts for map (stokes={stokes}, nu={Quantity(nu, 'Hz')})")

        data = map_data["sum"] / map_data["wgt"]

        # by convention maps have zero mean
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            data_offsets = np.nanmean(data, axis=(-1, -2))[..., None, None]

        beam_sum = np.zeros((len(self.bands), 3))
        beam_wgt = np.zeros((len(self.bands), 3))

        for band_index, band in enumerate(self.bands):
            for tod in self.tods:
                beam_sum[band_index] += tod.duration * tod.dets.beams[tod.dets.band_name == band.name].mean(axis=0)
                beam_wgt[band_index] += tod.duration

        return ProjectedMap(
            data=(data - data_offsets),
            stokes=self.stokes,
            weight=map_data["wgt"],
            nu=self.bands.center,
            resolution=self.resolution,
            center=self.center,
            degrees=False,
            frame=self.frame.name,
            units=self.units,
            beam=beam_sum / beam_wgt,
        )



from __future__ import annotations

from collections.abc import Mapping, Sequence

import matplotlib.pyplot as plt
import pandas as pd

from .band import Band, parse_band


class BandList(Sequence):
    def __init__(self, bands: Mapping | list = []):
        self.bands = []

        if isinstance(bands, BandList):
            for band in bands.bands:
                self.add(parse_band(band))

        elif isinstance(bands, Mapping):
            for band_name, band in bands.items():
                b = parse_band(band)
                b.name = band_name
                self.add(b)

        elif isinstance(bands, list):
            for band in bands:
                self.add(parse_band(band))

    @property
    def nu_min(self):
        return min([band.nu.min() for band in self.bands])

    @property
    def nu_max(self):
        return max([band.nu.max() for band in self.bands])

    def plot(self):
        for band in self.bands:
            fig, ax = plt.subplots(1, 1)
            ax.plot(band.nu.Hz, band.tau, label=band.name)
        ax.set_xlim(self.nu_min.Hz, self.nu_max.Hz)
        ax.set_xlabel(r"$\nu$ [GHz]")
        ax.set_ylabel(r"$\tau(\nu)$ [Rayleigh-Jeans]")
        ax.legend()

    def add(self, band):
        if not isinstance(band, Band):
            raise ValueError("'band' must be a Band type.")
        if band.name in self.names:
            self.bands[self.names.index(band.name)] = band
        else:
            self.bands.append(band)

    def __getattr__(self, attr):
        if attr in self.names:
            return self.__getitem__(attr)
        if all([hasattr(band, attr) for band in self.bands]):
            return [getattr(band, attr) for band in self.bands]
        raise AttributeError(f"'BandList' object has no attribute '{attr}'")

    def __getitem__(self, index):
        if type(index) is int:
            return self.bands[index]
        elif type(index) is str:
            if index not in self.names:
                raise ValueError(f"BandList has no band named {index}.")
            return self.bands[self.names.index(index)]
        else:
            raise ValueError(
                f"Invalid index {index}. A bandList must be indexed by either an integer or a string.",
            )

    def __len__(self):
        return len(self.bands)

    def __repr__(self):
        return self.summary().__repr__()

    def _repr_html_(self):
        return self.summary()._repr_html_()

    def __short_repr__(self):
        return f"BandList([{', '.join(self.names)}])"

    @property
    def names(self):
        return [band.name for band in self.bands]

    def summary(self) -> pd.DataFrame:
        return pd.concat([band.summary() for band in self.bands], axis=1).T



from __future__ import annotations

import numpy as np
import scipy as sp


def weighted_binned_mean(x, y, bins, ignore_nan=True, weights=None):
    if weights is None:
        weights = np.ones(x.shape)

    numer = sp.stats.binned_statistic(x, weights * y, bins=bins, statistic="sum")[0]
    denom = sp.stats.binned_statistic(x, weights, bins=bins, statistic="sum")[0]

    return numer / denom


def get_kernel(n, kind="triangle"):
    if kind == "triangle":
        kernel = np.r_[np.linspace(0, 1, n + 1)[1:], np.linspace(1, 0, n + 1)[1:-1]]

        return kernel / kernel.sum()


def unwrap_angle(angle):
    mean_angle = np.angle(np.exp(1j * angle).mean())
    adju_angle = np.angle(np.exp(1j * (angle - mean_angle))) + mean_angle
    cntr_angle = 0.5 * (adju_angle.max() + adju_angle.min())
    return np.angle(np.exp(1j * (angle - cntr_angle))) + cntr_angle


def downsample(DATA, rate, axis=-1, method="triangle"):
    if method == "flat":
        _DATA = np.swapaxes(DATA, 0, axis)
        cs_data = np.cumsum(_DATA, axis=0)
        return np.swapaxes((cs_data[rate::rate] - cs_data[:-rate:rate]) / rate, 0, axis)

    else:
        if rate == 1:
            return DATA
        if rate < 1:
            raise ValueError("downsample rate must be an integer >= 1")

        _DATA = np.swapaxes(DATA, 0, axis)
        kernel = np.expand_dims(
            get_kernel(n=rate, kind=method),
            axis=tuple(np.arange(1, len(DATA.shape))),
        )
        n_kern = len(kernel)
        starts = np.arange(0, len(_DATA) - n_kern, rate)
        ends = starts + n_kern

        return np.swapaxes(
            np.r_[[np.sum(_DATA[s:e] * kernel, axis=0) for s, e in zip(starts, ends)]],
            0,
            axis,
        )


def get_phase_template(DATA, phase, n_phase_bins, discriminator=None):
    if discriminator is None:
        discriminator = np.ones(DATA.shape[0])

    nd, nt = DATA.shape
    TEMPLATE = np.zeros((nd, nt))

    for ud in np.unique(discriminator):
        mask = discriminator == ud
        D = DATA[mask].copy()
        D_mean = D.mean(axis=0)

        fractional_bin_index = phase * (n_phase_bins / (2 * np.pi))

        from sklearn.preprocessing import PolynomialFeatures

        template_degree = 2
        poly = PolynomialFeatures(degree=template_degree).fit_transform(
            np.linspace(-1, 1, nt)[:, None],
        )

        P = np.zeros((nt, n_phase_bins))
        P[np.arange(nt), np.floor(fractional_bin_index).astype(int) % n_phase_bins] = 1 - fractional_bin_index % 1
        P[np.arange(nt), np.ceil(fractional_bin_index).astype(int) % n_phase_bins] = fractional_bin_index % 1

        P = sp.ndimage.gaussian_filter1d(P, sigma=1, axis=1, mode="wrap")
        PP = np.concatenate(
            [P * poly[:, i][:, None] for i in range(template_degree + 1)],
            axis=1,
        )
        PD = np.matmul(np.linalg.inv(np.matmul(PP.T, PP)), np.matmul(PP.T, D_mean))
        template = np.matmul(PP, PD)

        gains = np.sum(template * D, axis=1) / np.square(template).sum()
        TEMPLATE[mask] = np.outer(gains, template)

    return TEMPLATE


def make_cuts(D, n_filt=3, downsample_rate=4, max_cuts=256):
    ds_D = downsample(D, rate=downsample_rate, method="triangle")

    filt = -np.ones(n_filt) / (n_filt - 1)
    filt[int((n_filt - 1) / 2)] = 1

    residual = sp.ndimage.convolve1d(ds_D, filt, axis=1)
    mnd, mnt = D.shape
    cuts = []

    for i, _res in enumerate(residual):
        cuts.append([])
        sq_res = np.square(_res)
        med = np.median(sq_res[::4])

        is_bad = (sq_res > 1e2 * med) | np.isnan(sq_res)

        for sub_index in grouper(np.where(is_bad)[0], tol=2):
            s, e = np.min(sub_index) - 1, np.max(sub_index) + 1
            if s < 0 or e > mnt - 1:
                continue
            # sub_res_sum = _res[sub_index].sum()
            cuts[-1].append(
                (
                    downsample_rate * np.min(sub_index) - 1,
                    downsample_rate * np.max(sub_index) + 1,
                ),
            )

        if len(cuts[-1]) > max_cuts:
            cuts[-1] = [(0, mnt - 1)]

    return cuts


def apply_cuts(D, cuts, tol=4, method=None):
    fD, T = D.copy(), D.shape[1]
    for i, _cuts in enumerate(cuts):
        for s, e in _cuts:
            if e - s > 1024:
                fD[i, 0] = np.nan
                continue
            elif method == "splice":
                t0, t1 = np.maximum(s - 1, 0), np.minimum(e, T - 1)
                fD[i, t0:t1] = np.linspace(fD[i, t0], fD[i, t1], t1 - t0)
            elif method == "flatten":
                i0, i1, i2, i3 = (
                    np.maximum(s - tol, 0),
                    s,
                    e,
                    np.minimum(e + tol, T - 1),
                )
                if not i0 < i1 < i2 < i3:
                    continue
                d0, d1 = np.median(fD[i, i0:i1]), np.median(fD[i, i2:i3])
                fD[i, i2:] -= d1 - d0
                fD[i, i1:i2] = np.linspace(d0, d0, i2 - i1)
    return fD


def decompose(DATA, mode="us", downsample_rate=1):
    downsampled_data = downsample(DATA, rate=downsample_rate, method="triangle")
    rms = downsampled_data.std(axis=-1)
    u, s, v = np.linalg.svd(downsampled_data / rms[:, None], full_matrices=False)
    uv_norm = v.std(axis=-1) * np.sign(u.mean(axis=0))
    s_norm = np.sqrt(np.square(s).sum())
    u *= s_norm * rms[:, None] * uv_norm[None, :]
    s /= s_norm

    if mode == "us":
        return u, s
    if mode == "uv":
        return np.matmul(u, np.diag(s)), np.matmul(
            np.linalg.pinv(np.matmul(u, np.diag(s))),
            DATA,
        )
    if mode == "usv":
        return u, s, np.matmul(np.linalg.pinv(np.matmul(u, np.diag(s))), DATA)


def bandpass(data, lc, hc, fs, order):
    return sp.signal.filtfilt(
        *sp.signal.butter(order, [2 * lc / fs, 2 * hc / fs], btype="band"),
        data,
        axis=-1,
    )


def lowpass(data, c, fs, order):
    return sp.signal.filtfilt(
        *sp.signal.butter(order, 2 * c / fs, btype="lowpass"),
        data,
        axis=-1,
    )


def highpass(data, c, fs, order):
    return sp.signal.filtfilt(
        *sp.signal.butter(order, 2 * c / fs, btype="highpass"),
        data,
        axis=-1,
    )



import numpy as np
from dask import array as da


class Field:
    def __init__(self, data: float, dtype: type = np.float32):
        self.dtype = dtype
        self.data = data

    @property
    def data(self):
        return self._data + self._offset[..., None]

    @data.setter
    def data(self, value):
        self._offset = value.mean(axis=-1)  # average over time axis
        self._data = da.asarray(value - self._offset[..., None]).astype(self.dtype)

    def __getitem__(self, key):
        return Field(data=self.data.__getitem__(key), dtype=self.dtype)

    def __repr__(self):
        return f"Field({self.data.__repr__()})"



from __future__ import annotations

import logging
import time as ttime

import dask.array as da
import healpy as hp
import numpy as np
import scipy as sp
from tqdm import tqdm

from ..cmb import CMB, DEFAULT_CMB_KWARGS, generate_cmb, get_cmb
from ..constants import T_CMB, k_B
from ..functions.radiometry import (
    inverse_planck_spectrum,
    inverse_rayleigh_jeans_spectrum,
    planck_spectrum,
    rayleigh_jeans_spectrum,
)  # noqa
from ..io import DEFAULT_BAR_FORMAT, humanize_time
from .observation import Observation

logger = logging.getLogger("maria")


class CMBMixin:
    def _init_cmb(self, cmb: str | CMB, **cmb_kwargs):
        self.cmb_kwargs = DEFAULT_CMB_KWARGS.copy()
        self.cmb_kwargs.update(cmb_kwargs)

        if cmb in ["spectrum", "power_spectrum", "generate", "generated"]:
            for _ in tqdm(
                range(1),
                desc=f"Generating CMB (nside={self.cmb_kwargs['nside']})",
                disable=self.disable_progress_bars,
            ):
                self.cmb = generate_cmb(**self.cmb_kwargs)
        elif cmb in ["real", "planck"]:
            self.cmb = get_cmb(**self.cmb_kwargs)
        else:
            raise ValueError(f"Invalid value for cmb '{cmb}'.")

    def _compute_cmb_loading(self, obs: Observation, eps: float = 1e-6):
        cmb_loading = np.zeros(obs.shape, dtype=self.dtype)

        bands_pbar = tqdm(
            obs.instrument.bands,
            desc="Sampling CMB",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
            ncols=250,
        )

        stokes_weight = obs.instrument.dets.stokes_weight()

        for band in bands_pbar:
            bands_pbar.set_postfix(band=band.name)
            band_mask = obs.instrument.dets.band_name == band.name

            # the CMB is not a Rayleigh-Jeans source, so we do the power integrals below
            test_T_b = np.array([T_CMB, T_CMB + eps])
            test_T_RJ = inverse_rayleigh_jeans_spectrum(
                planck_spectrum(T_b=test_T_b, nu=band.nu.Hz[:, None]), nu=band.nu.Hz[:, None]
            )

            if hasattr(obs, "atmosphere"):
                opacity = sp.interpolate.interp1d(obs.atmosphere.spectrum.side_nu, obs.atmosphere.spectrum._opacity)(
                    band.nu.Hz
                )

                # the 1e12 is for picowatts
                det_power_grid = (
                    1e12
                    * k_B
                    * np.trapezoid(
                        test_T_RJ[None, None, None] * (np.exp(-opacity) * band.passband(band.nu.Hz))[..., None],
                        x=band.nu.Hz,
                        axis=-2,
                    )
                )

                P = sp.interpolate.RegularGridInterpolator(
                    obs.atmosphere.spectrum.points[:3],
                    det_power_grid,
                )((obs.atmosphere.weather.temperature[0], obs.zenith_scaled_pwv[band_mask], obs.coords.el[band_mask]))

            else:
                P = (
                    1e12
                    * k_B
                    * np.trapezoid(
                        test_T_RJ * band.passband(band.nu.Hz[:, None]),
                        x=band.nu.Hz,
                        axis=-2,
                    )
                )

            assert P.shape[-1] == 2

            pW_per_K_CMB = (P[..., 1] - P[..., 0]) / eps

            # flat_band_pixel_index = hp.ang2pix(
            #     nside=obs.cmb.nside,
            #     phi=obs.coords.l[band_mask],
            #     theta=np.pi / 2 - obs.coords.b[band_mask],
            # ).ravel()

            band_coords = obs.coords[band_mask]

            pointing_s = ttime.monotonic()
            pmat = self.cmb.pointing_matrix(band_coords)
            logger.debug(
                f"Computed CMB pointing matrix for band {band.name} in {humanize_time(ttime.monotonic() - pointing_s)}"
            )

            # the total loading from the mean temperature
            loading_s = ttime.monotonic()
            cmb_loading[band_mask] = P[..., 0] * stokes_weight[band_mask, 0][:, None]
            logger.debug(f"Computed CMB loading for band {band.name} in {humanize_time(ttime.monotonic() - loading_s)}")

            for stokes_index, stokes in enumerate("IQU"):
                bands_pbar.set_postfix(band=band.name, stokes=stokes)
                cmb_stokes_s = ttime.monotonic()
                s = stokes_weight[band_mask, stokes_index][:, None]
                if np.isclose(s, 0).all():
                    logger.debug(f"Skipping stokes {stokes} for CMB (no weight)")
                    continue

                cmb_loading[band_mask] += (
                    pW_per_K_CMB * s * (self.cmb.data[stokes_index, 0].compute() @ pmat).reshape(band_coords.shape)
                )
                logger.debug(
                    f"Computed Stokes {stokes} CMB anisotropy for band {band.name} "
                    f"in {humanize_time(ttime.monotonic() - cmb_stokes_s)}"
                )

        return da.asarray(cmb_loading, dtype=self.dtype)



from ..map import HEALPixMap


class CMB(HEALPixMap):
    def __init__(
        self,
        data: float,
        weight: float = None,
        stokes: float = None,
        frame: str = "galactic",
        nu: float = None,
        units: str = "K_CMB",
    ):
        super().__init__(data=data, weight=weight, stokes=stokes, nu=nu, z=1100.0, units=units, frame=frame)



from __future__ import annotations

import numpy as np

from maria.coords import offsets_to_phi_theta, phi_theta_to_offsets


def test_offsets_transform():
    n = 256

    for cphi in np.random.uniform(low=0, high=2 * np.pi, size=5):
        for ctheta in np.random.uniform(low=-np.pi / 2, high=np.pi / 2, size=5):
            offsets = np.radians(
                np.random.uniform(low=-0.5, high=+0.5, size=(n, 2)),
            )

            _phitheta = offsets_to_phi_theta(offsets, cphi, ctheta)
            _offsets = phi_theta_to_offsets(_phitheta, cphi, ctheta)

            assert np.mean(np.square(offsets - _offsets)) < 1e-5



from __future__ import annotations

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np


def unjitted_offsets_to_phi_theta(dX, cphi, ctheta):
    """
    A fast and well-conditioned method to convert from local dx/dy coordinates to phi/theta coordinates.
    """

    dx, dy = dX[..., 0], dX[..., 1]

    r = jnp.sqrt(dx**2 + dy**2)  # distance from the center
    p = jnp.arctan2(dx, -dy)  # 0 at the bottom, increases CCW to pi at the top

    # if we're looking at the north pole, we have (lon, lat) = (p, pi/2 - r)
    # a projection looking from the east
    proj_from_east = (jnp.sin(r) * jnp.cos(p) + 1j * jnp.cos(r)) * jnp.exp(
        1j * (ctheta - jnp.pi / 2),
    )

    return jnp.stack(
        [jnp.arctan2(jnp.sin(r) * jnp.sin(p), jnp.real(proj_from_east)) + cphi, jnp.arcsin(jnp.imag(proj_from_east))],
        axis=-1,
    )


offsets_to_phi_theta = jax.jit(unjitted_offsets_to_phi_theta, static_argnames=["cphi", "ctheta"])


@partial(jax.jit, static_argnames=["cphi", "ctheta"])
def phi_theta_to_offsets(pt, cphi, ctheta):
    """
    A fast and well-conditioned to convert from phi/theta coordinates to local dx/dy coordinates.
    """

    phi, theta = pt[..., 0], pt[..., 1]

    dphi = phi - cphi
    proj_from_east = (jnp.cos(dphi) * jnp.cos(theta) + 1j * jnp.sin(theta)) * jnp.exp(
        1j * (jnp.pi / 2 - ctheta),
    )
    dz = jnp.sin(dphi) * jnp.cos(theta) + 1j * jnp.real(proj_from_east)
    r = jnp.abs(dz)
    dz *= jnp.arcsin(r) / jnp.where(r > 0, r, 1.0)

    # negative, because we're looking at the observer
    return jnp.stack([jnp.real(dz), -jnp.imag(dz)], axis=-1)


@jax.jit
def phi_theta_to_xyz(phi, theta):
    """
    Project some angular coordinates phi (longitude) and theta (latitude) onto the 3d unit sphere.
    """
    cos_theta = jnp.cos(theta)
    return jnp.stack(
        [jnp.cos(phi) * cos_theta, jnp.sin(phi) * cos_theta, jnp.sin(theta)],
        axis=-1,
    )


@jax.jit
def xyz_to_phi_theta(xyz):
    """
    Find the longitude and latitude of a 3-vector.
    """
    return jnp.arctan2(xyz[..., 1], xyz[..., 0]) % (2 * jnp.pi), jnp.arcsin(
        xyz[..., 2] / jnp.sqrt(jnp.sum(xyz**2, axis=-1)),
    )


@partial(jax.jit, static_argnames=["keep_dims"])
def get_center_phi_theta(phi, theta, keep_dims=()):
    """ """

    xyz = phi_theta_to_xyz(jnp.atleast_1d(phi), jnp.atleast_1d(theta))

    axes = list(range(xyz.ndim - 1))

    for dim in keep_dims:
        axes.pop(dim)

    center_xyz = xyz.mean(axis=tuple(axes))
    center_xyz /= jnp.sqrt(jnp.sum(jnp.square(center_xyz)))[..., None]

    return xyz_to_phi_theta(center_xyz)



from __future__ import annotations

import logging
import os

from ..spectrum import AtmosphericSpectrum
from .atmosphere import Atmosphere

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")

SUPPORTED_MODELS_LIST = ["2d", "3d"]

DEFAULT_ATMOSPHERE_KWARGS = {}



from __future__ import annotations

import numpy as np
import pytest

from maria.units import Quantity


def test_units_repr():
    assert str(Quantity(0.1, "deg")) == "6’"
    assert str(Quantity(1e-8, "deg")) == "36 uarcsec"
    assert str(Quantity(1e6, "uK_RJ")) == "1 K_RJ"
    assert str(Quantity(1e15, "K_RJ")) == "1e+15 K_RJ"  # don't go high
    assert str(Quantity(1e-6, "Hz")) == "1e-06 Hz"  # don't go low
    assert str(Quantity(1000, "s")) == "1000 s"


def test_units_ops():
    deg = Quantity(1, "deg")
    rad = Quantity(1, "rad")
    sr = Quantity(1, "sr")
    sqdeg = Quantity(1, "deg^2")
    W = Quantity(1, "W")
    J = Quantity(1, "J")
    s = Quantity(1, "s")

    assert deg * deg == sqdeg
    assert rad * rad == sr
    assert rad / deg == np.degrees(1)
    assert J / s == W


def test_casting():
    for nu in [
        (90e9, 150e9, 220e9),
        np.array([90e9, 150e9, 220e9]),
        [Quantity(90, "GHz"), Quantity(150, "GHz"), Quantity(220, "GHz")],
        Quantity((90e9, 150e9, 220e9), "Hz"),
    ]:
        q = Quantity(nu, "Hz")
        assert all(q.GHz == [90, 150, 220])



from __future__ import annotations

from .coordinates import Coordinates, frames  # noqa
from .frame import FRAMES, Frame  # noqa
from .transforms import (  # noqa
    get_center_phi_theta,
    offsets_to_phi_theta,
    phi_theta_to_offsets,
    phi_theta_to_xyz,
    unjitted_offsets_to_phi_theta,
    xyz_to_phi_theta,
)



import logging
import os
import time as ttime
from collections.abc import Iterable

import arrow
import dask.array as da
import numpy as np
import scipy as sp

from ..calibration import Calibration
from ..constants import MARIA_MAX_NU_HZ, MARIA_MIN_NU_HZ
from ..errors import FrequencyOutOfBoundsError, ShapeError
from ..units import Quantity, parse_units

logger = logging.getLogger("maria")

here, this_filename = os.path.split(__file__)

SLICE_DIMS = {
    "stokes": {
        "dtype": str,
        "default": "I",
    },
    "nu": {
        "dtype": float,
        "default": 150e9,
    },
    "t": {
        "dtype": float,
        "default": arrow.now().timestamp(),
    },
}


class Map:
    """
    The base class for maps. Maps have data
    """

    def __init__(
        self,
        data: float,
        weight: float,
        stokes: str,
        nu: Iterable[float],
        t: Iterable[float],
        z: Iterable[float],
        beam: tuple[float, float, float],  # noqa
        map_dims: dict,
        units: str = "K_RJ",
        degrees: bool = True,  # noqa
        dtype: type = np.float32,
    ):
        self.data = da.asarray(data).astype(dtype)
        self.weight = (da.asarray(weight) if weight is not None else da.ones_like(self.data)).astype(dtype)
        self.data, self.weight = np.broadcast_arrays(self.data, self.weight)

        self.units = parse_units(units)["units"]

        if not hasattr(beam, "__len__"):
            beam = (beam, beam, 0)
        self.beam = Quantity(beam, "deg" if degrees else "rad")

        if np.shape(self.beam)[-1] != 3:
            raise ValueError("'beam' must be either a number or a tuple of (major, minor, angle)")

        self.map_dims = map_dims

        self.slice_dims = {}
        if stokes is not None:
            stokes = "".join(list(stokes))
            if stokes not in ["I", "IQU", "IQUV"]:
                raise ValueError(f"Invalid stokes parameter '{stokes}' (must be either 'I', 'IQU', or 'IQUV').")
            self.stokes = stokes.upper()
            self.slice_dims["stokes"] = len(self.stokes)
        else:
            self.stokes = ""

        if nu is not None:
            self.nu = Quantity(nu, "Hz")
            if self.nu.ndim == 0:
                self.nu = Quantity([self.nu.Hz], "Hz")
            if self.nu.ndim > 1:
                raise ShapeError("'nu' can be at most one-dimensional")
            bad_freqs = list(self.nu[(self.nu.Hz < MARIA_MIN_NU_HZ) | (self.nu.Hz > MARIA_MAX_NU_HZ)])
            if bad_freqs:
                raise FrequencyOutOfBoundsError(bad_freqs)
            self.slice_dims["nu"] = len(self.nu)
        else:
            self.nu = Quantity([], "Hz")

        if t is not None:
            self.t = np.atleast_1d(t)
            self.slice_dims["t"] = len(self.t)
        else:
            self.t = np.array([])

        if z is not None:
            self.z = np.atleast_1d(z)
            self.slice_dims["z"] = len(self.z)
        else:
            self.z = np.array([])

        self.data *= np.ones(list(self.dims.values()))
        self.weight *= np.ones(list(self.dims.values()))

        # for i, (dim, n) in enumerate(slice_dims.items()):
        #     if self.data.shape[i] != n:
        #         raise ValueError(
        #             f"'{dim}' has length {n} but map has {dim} length {self.data.shape[i]}.",
        #         )

        implied_shape = tuple(list(self.dims.values()))
        # if len(implied_shape) != self.data.ndim:
        #     raise ValueError(f"Inputs imply rank {len(self.dims)} for map data, but data has rank {self.data.ndim}")

        if implied_shape != self.data.shape:
            raise ValueError(
                f"Inputs imply shape {self.dims_string}={implied_shape} for map data, but data has shape {self.data.shape}"
            )

        if self.u["quantity"] == "spectral_flux_density_per_beam":
            if not np.all(self.beam_area > 0):
                raise ValueError(
                    f"Map is given in units {self.units}, but specified beam(major, minor, angle) = {beam} has zero area"
                )

    @property
    def dims(self):
        return {**self.slice_dims, **self.map_dims}

    @property
    def ndim(self):
        return len(self.dims)

    @property
    def shape(self):
        return tuple(self.dims.values())

    @property
    def dims_string(self):
        return f"({', '.join(self.dims.keys())})"

    @property
    def dims_list(self):
        return list(self.dims.keys())

    @property
    def nu_bins(self):
        return np.array([0, *(self.nu.Hz[1:] + self.nu.Hz[:-1]), 1e6])

    @property
    def nu_side(self):
        return (self.nu_bins[:-1] + self.nu_bins[1:]) / 2

    @property
    def t_bins(self):
        """
        This might break in the year 3000.
        """
        t_min = arrow.get("0001-01-01").timestamp()
        t_max = arrow.get("3000-01-01").timestamp()

        return np.array([t_min, *(self.t[1:] + self.t[:-1]), t_max])

    @property
    def t_side(self):
        return (self.t_bins[:-1] + self.t_bins[1:]) / 2

    @property
    def u(self):
        return parse_units(self.units)

    # @property
    # def weight(self):
    #     return self._weight if self._weight is not None else da.ones(shape=self.data.shape)

    def append(self, map, dim):
        return concatenate([self, map], dim=dim)

    def extend(self, maps, dim):
        return concatenate([self, *maps], dim=dim)

    def squeeze(self, dims=None):
        dims = np.atleast_1d(dims) if dims is not None else [dim for dim, n in self.slice_dims.items() if n == 1]
        package = self.package()
        dim_indices_to_squeeze = []

        for dim_index, name in enumerate(dims):
            n = self.slice_dims.get(name)
            if n is None:
                raise ValueError(f"{self.__class__.__name__} has no dimension '{name}'")
            if n != 1:
                raise ValueError(f"Cannot squeeze dimension '{name}' with length {n} > 1")

            dim_indices_to_squeeze.append(dim_index)
            package.pop(name)

        package["data"] = package["data"].squeeze(tuple(dim_indices_to_squeeze))
        package["weight"] = package["weight"].squeeze(tuple(dim_indices_to_squeeze))

        return type(self)(**package)

    def unsqueeze(self, dim, value=None):
        if dim not in SLICE_DIMS:
            raise ValueError(f"'{dim}' is not a valid map dimension")
        if dim in self.dims:
            raise Exception(f"{self.__class__.__name__} already has dimension '{dim}'")

        new_dim_index = 0
        for d in ["stokes", "nu", "t"]:
            if d == dim:
                break
            if d in self.dims:
                new_dim_index += 1

        package = self.package()
        package["data"] = np.expand_dims(package["data"], new_dim_index)
        package["weight"] = np.expand_dims(package["weight"], new_dim_index)
        if value is None:
            value = SLICE_DIMS[dim]["default"]
        package[dim] = value

        return type(self)(**package)

    @property
    def pixel_area(self):
        if hasattr(self, "resolution"):
            return Quantity(self.resolution.rad**2, "sr")
        else:
            return Quantity(self.x_res.rad * self.y_res.rad, "sr")

    @property
    def beam_area(self):
        """
        Returns the beam area in steradians
        """
        area = (np.pi / 4) * self.beam[..., 0].radians * self.beam[..., 1].radians
        return Quantity(area * np.ones(tuple(self.slice_dims.values())), "sr")

    def to(self, units: str):
        if units == self.units:
            return self

        u = parse_units(units)

        package = self.package().copy()

        # this is just a scaling by some factor
        if u["quantity"] == self.u["quantity"]:
            package["data"] *= self.u["factor"] / u["factor"]

        else:
            if "nu" not in self.dims:
                raise ValueError(
                    f"Cannot convert from quantity {self.u['quantity']} to quantity {u['quantity']} "
                    f"when map has no frequency."
                )

            # data = package["data"].swapaxes(0, self.dims_list.index("nu"))  # put the nu index in front
            for nu_index, nu in enumerate(self.nu.Hz):
                nu_key = tuple([nu_index if dim == "nu" else slice(None, None, None) for dim in self.slice_dims])

                cal = Calibration(
                    f"{self.units} -> {units}",
                    nu=nu,
                    pixel_area=self.pixel_area.sr,
                    beam_area=self.beam_area[nu_key].sr,
                )
                package["data"][nu_key] = cal(package["data"][nu_key])

            # package["data"] = data.swapaxes(0, self.dims_list.index("nu"))  # swap the axes back

        package["units"] = units
        return type(self)(**package)

    def sample_nu(self, nu):
        map_nu_interpolator = sp.interpolate.interp1d(self.nu.Hz, self.data, axis=1, kind="linear")

        nu_maps = []
        for nu in np.atleast_1d(nu):
            if nu < self.nu.Hz[0]:
                nu_maps.append(self.data[:, 0])
            elif not (nu < self.nu.Hz[-1]):  # this will include nan
                nu_maps.append(self.data[:, -1])
            else:
                nu_maps.append(map_nu_interpolator(nu))

        return np.stack(nu_maps, axis=1)

    @property
    def nu_bin_bounds(self):
        nu_boundaries = [0, *(self.nu.Hz[:-1] + self.nu.Hz[1:]) / 2, np.inf]
        return [(Quantity(nu1, "Hz"), Quantity(nu2, "Hz")) for nu1, nu2 in zip(nu_boundaries[:-1], nu_boundaries[1:])]

    def copy(self):
        return type(self)(**self.package().copy())


def concatenate(maps: list[Map], dim: str) -> Map:
    packages = []
    concat_dim_values = []
    dims_list = {}
    for m in maps:
        for d, n in m.dims.items():
            if d not in dims_list:
                dims_list[d] = []
            dims_list[d].append(n)
        for attr in ["center", "width", "height"]:
            # TODO: checks for healpix maps
            if getattr(m, attr, None) != getattr(maps[0], attr, None):
                print("warning")
        concat_dim_values.extend(getattr(m, dim))
        packages.append(m.to(maps[0].units).package())

    for d, ns in dims_list.items():
        if d != dim:
            if len(set(ns)) > 1:
                raise ShapeError(
                    "Map dimensions must be equal except along the concatenation axis "
                    f"(maps have shapes {[m.shape for m in maps]})."
                )

    dim_index = list(dims_list.keys()).index(dim)
    total_package = packages[0].copy()
    total_package["data"] = np.concatenate([p["data"] for p in packages], axis=dim_index)
    total_package["weight"] = np.concatenate([p["weight"] for p in packages], axis=dim_index)
    total_package[dim] = concat_dim_values

    return type(maps[0])(**total_package)



from __future__ import annotations

import numpy as np
import scipy as sp

from .filters import bandpass, highpass, lowpass  # noqa


def get_kernel(n, kind="triangle"):
    if kind == "triangle":
        kernel = np.r_[np.linspace(0, 1, n + 1)[1:], np.linspace(1, 0, n + 1)[1:-1]]

        return kernel / kernel.sum()


def fast_downsample(DATA, r=1):
    *in_shape, n = DATA.shape
    CUMSUM = np.cumsum(np.atleast_2d(DATA), axis=-1)
    return (CUMSUM[..., r::r] - CUMSUM[..., :-r:r]).reshape(*in_shape, -1) / r


def downsample(DATA, rate, axis=-1, method=None):
    if method is None:
        return np.swapaxes(np.swapaxes(DATA, 0, axis)[::rate], 0, axis)

    if method == "fast":
        return fast_downsample(
            DATA,
            r=rate,
        )

    if method == "flat":
        _DATA = np.swapaxes(DATA, 0, axis)
        cs_data = np.cumsum(_DATA, axis=0)
        return np.swapaxes((cs_data[rate::rate] - cs_data[:-rate:rate]) / rate, 0, axis)

    if method == "triangle":
        if rate == 1:
            return DATA
        if rate < 1:
            raise ValueError("downsample rate must be an integer >= 1")

        _DATA = np.swapaxes(DATA, 0, axis)
        kernel = np.expand_dims(
            get_kernel(n=rate, kind="triangle"),
            axis=tuple(np.arange(1, len(DATA.shape))),
        )
        n_kern = len(kernel)
        starts = np.arange(0, len(_DATA) - n_kern, rate)
        ends = starts + n_kern

        return np.swapaxes(
            np.r_[[np.sum(_DATA[s:e] * kernel, axis=0) for s, e in zip(starts, ends)]],
            0,
            axis,
        )


def decompose(data, k: int = None):
    k = k or len(data)

    if data.ndim != 2:
        raise ValueError("'data' must be a 2D array.")

    if k < len(data):
        u, s, vh = sp.sparse.linalg.svds(data, k=k, which="LM")
    else:
        u, s, vh = sp.linalg.svd(data, full_matrices=False)

    norm = np.sign(u.mean(axis=0)) * vh.std(axis=-1)
    a = u @ np.diag(s * norm)
    b = np.diag(1 / norm) @ vh

    mode_sort = np.argsort(-s)

    return a[:, mode_sort], b[mode_sort]


def bspline_knots(t, spacing, order):
    tmin = t.min()
    tmax = t.max()
    n_bins = int(np.maximum((tmax - tmin) // spacing, 1))  # how many bins straddle the data?
    n_basis = n_bins + order + 1

    k = spacing * np.arange(n_bins, dtype=float)  # so that it straddles the whole domain
    k += float(tmax + tmin) / 2 - k.mean()
    k = np.r_[
        k[0] + spacing * np.arange(-order - 1, 0),
        k,
        k[-1] + spacing * np.arange(1, order + 2),
    ]  # pad the edges

    return k


def bspline_basis_from_knots(t, k, order):
    n_basis = len(k) - order - 1
    B = np.zeros((len(k) + 1, order + 1, len(t)))
    B[np.digitize(t, k) - 1, 0, np.arange(len(t))] = 1

    for p in range(1, order + 1):
        for i in range(len(k) - p - 1):
            B[i, p] = B[i, p - 1] * (t - k[i]) / (k[i + p] - k[i]) + B[i + 1, p - 1] * (k[i + p + 1] - t) / (
                k[i + p + 1] - k[i + 1]
            )

    return B[:n_basis, -1]


def bspline_basis(t, spacing, order):
    k = bspline_knots(t, spacing, order)
    return bspline_basis_from_knots(t, k, order)


def fit_bspline(y, x, spacing, order=3):
    B = bspline_basis(x, spacing=spacing, order=order)
    A = y @ (np.linalg.inv(B @ B.T) @ B).T
    return A @ B


def cross_basis(X: list, spacing: list, order: list):
    basis = np.ones((1, 1))

    for dim, x in enumerate(X):
        x_basis = bspline_basis(x, spacing[dim], order[dim])
        basis = (x_basis[:, None] * basis).reshape(-1, len(x))
        basis = basis[basis.sum(axis=-1) > 0]

    return basis


def detrend(D, order=3):
    x = np.linspace(-1, 1, D.shape[-1])
    X = np.c_[[x**i for i in range(order + 1)]]
    A = D @ X.T @ np.linalg.inv(X @ X.T).T

    return D - A @ X


def remove_slope(D):
    return D - np.linspace(D[..., 0], D[..., -1], D.shape[-1]).T


def grouper(iterable, min_length=1, max_length=np.inf, overlap=False):
    start = np.inf
    prev_value = False
    for index, this_value in enumerate(iterable):
        if this_value:
            if not prev_value:
                start = index
            if prev_value:
                if index - start >= max_length:
                    yield (start, index)
                    start = index
        if not this_value:
            if prev_value:
                if index - start >= min_length:
                    yield (start, index)
        prev_value = this_value

    if prev_value:
        yield (start, index + 1)


# print(list(grouper([True, True, False, True, True])))
# print(list(grouper([False, True, True, True, True, True, False, True, True], max_length=10, overlap=True)))
# print(list(grouper([False, True, True, True, True, True, False, True, True], max_length=10, overlap=False)))



from __future__ import annotations

import logging

import healpy as hp
import numpy as np
import pandas as pd

from ..io import download_from_url, fetch
from .cmb import CMB

# shut up healpy I don't care about the resolution
logging.getLogger("healpy").setLevel(logging.WARNING)

CMB_SPECTRUM_SOURCE_URL = (
    "https://github.com/thomaswmorris/maria-data/raw/master/cmb/spectra/"
    "COM_PowerSpect_CMB-base-plikHM-TTTEEE-lowl-lowE-lensing-minimum-theory_R3.01.txt"
)
CMB_SPECTRUM_CACHE_PATH = "/tmp/maria-data/cmb/spectrum.txt"
CMB_SPECTRUM_CACHE_MAX_AGE = 30 * 86400  # one month

CMB_MAP_SOURCE_URL = (
    "https://irsa.ipac.caltech.edu/data/Planck/release_3/all-sky-maps/maps/component-maps/cmb/"
    "COM_CMB_IQU-143-fgsub-sevem_2048_R3.00_full.fits"
)
CMB_MAP_CACHE_PATH = "/tmp/maria-data/cmb/planck/map.fits"
CMB_MAP_CACHE_MAX_AGE = 30 * 86400  # one month


DEFAULT_CMB_KWARGS = {"nside": 2048}


def get_cmb_spectrum():
    cmb_spectrum_path = fetch(
        "cmb/spectra/planck.csv",
        max_age=30 * 86400,
    )

    return pd.read_csv(cmb_spectrum_path, index_col=0)


def generate_cmb(nside=2048, seed=123456, **kwargs):
    """
    Generate a new CMB.

    Taken from https://www.zonca.dev/posts/2020-09-30-planck-spectra-healpy.html
    """

    cl = get_cmb_spectrum()
    lmax = cl.index.max()

    alm = hp.synalm((cl.TT, cl.EE, cl.BB, cl.TE), lmax=lmax, new=True)
    cmb_data = hp.alm2map(alm, nside=nside, lmax=lmax)

    return CMB(
        data=cmb_data[:, None, None, :],
        stokes="IQU",
        units="K_CMB",
        frame="galactic",
        nu=148e9,
    )


def get_cmb(**kwargs):
    download_from_url(
        source_url=CMB_MAP_SOURCE_URL,
        cache_path=CMB_MAP_CACHE_PATH,
    )

    field_dtypes = {
        "T": np.float32,
        "Q": np.float32,
        "U": np.float32,
        "T_mask": bool,
        "P_mask": bool,
    }

    maps = {
        field: hp.fitsfunc.read_map(CMB_MAP_CACHE_PATH, field=i).astype(dtype)
        for i, (field, dtype) in enumerate(field_dtypes.items())
    }

    maps["T"] = np.where(maps["T_MASK"], maps["T"], np.nan)
    maps["Q"] = np.where(maps["P_MASK"], maps["Q"], np.nan)
    maps["U"] = np.where(maps["P_MASK"], maps["U"], np.nan)

    return CMB(data=[maps["T"], maps["Q"], maps["U"]], fields=["T", "Q", "U"])



from __future__ import annotations

# nothing in here should import from other maria module (so no double dots)
import jax
import jax.numpy as jnp
import numpy as np
import scipy as sp

# from ..units.prefixes import *  # noqa
from .coords import *  # noqa
from .io import *  # noqa
from .linalg import *  # noqa
from .plotting import *  # noqa
from .signal import *  # noqa
from .time import *  # noqa


def hav(x):
    return (1 - np.cos(x)) / 2


def great_circle_distance(phi1, theta1, phi2, theta2):
    hav_d = hav(theta2 - theta1) + np.cos(theta1) * np.cos(theta2) * hav(phi1 - phi2)
    return 2 * np.sqrt(hav_d)


@jax.jit
def regular_digitization(x, bins):
    dx = jnp.gradient(bins).mean()
    return ((x - (bins.min() - dx)) / dx).astype(int).clip(min=0, max=len(bins))


def compute_pointing_matrix_sparse_indices(points, bins):
    if not np.all([np.all(np.gradient(b) > 0) for b in bins]):
        raise ValueError(f"Each set of bins must be strictly increasing")

    idx = 0
    cum_npix = 1
    for dim, (x, b) in enumerate(zip(points, bins)):
        idx += cum_npix * regular_digitization(x.ravel(), b)
        cum_npix *= len(b) + 1

    return idx, cum_npix


def generate_fourier_noise(nx: float = 1024, ny: float = 1024, k0: float = 5e0, beta: float = 8 / 3):
    kx = np.fft.fftfreq(nx, d=1 / nx)
    ky = np.fft.fftfreq(ny, d=1 / ny)
    KY, KX = np.meshgrid(ky, kx)
    P = np.sqrt(k0**2 + KX**2 + KY**2) ** (-beta - 1)
    F = np.fft.fft2(np.sqrt(P) * np.fft.ifft2(np.random.standard_normal(size=(ny, nx)))).real

    return (F - F.mean()) / F.std()


# def generate_power_law_noise(n: tuple = (256, 256), cutoff=1e0, beta=None):
#     ndim = len(n)  # noqa
#     beta = beta or 8 / 6

#     X_list = np.meshgrid(*[np.linspace(0, 1, _n) for _n in n])
#     K_list = np.meshgrid(*[np.fft.fftfreq(_n, d=1 / _n) for _n in n])

#     P = (cutoff**2 + sum([K**2 for K in K_list])) ** -(beta / 2)

#     F = np.real(np.fft.fftn(P * np.fft.ifftn(np.random.standard_normal(size=n))))

#     return *X_list, (F - F.mean()) / F.std()


def unpack_implicit_slice(key, ndims):
    explicit_slices = []
    for s in key:
        if s == Ellipsis:
            for _ in range(ndims + 1 - len(key)):
                explicit_slices.append(slice(None, None, None))
        else:
            explicit_slices.append(s)

    while len(explicit_slices) < ndims:
        explicit_slices.append(slice(None, None, None))

    return explicit_slices


def compute_diameter(points, lazy=False, MAX_SAMPLE_SIZE: int = 10000, jitter: float = 0.0) -> float:
    """
    Parameters
    ----------
    points : type
        An (..., n_dim) array of offsets of something.
    """

    *input_shape, n_dim = points.shape
    X = points.reshape(-1, n_dim)

    if jitter:
        X += jitter * np.ptp(X, axis=-1).max() * np.random.standard_normal(size=X.shape)

    dim_mask = np.ptp(X, axis=tuple(range(X.ndim - 1))) > 0
    if not dim_mask.any():
        return 0.0

    if len(X) < 16:
        i, j = np.triu_indices(n=len(X), k=1)
        return np.sqrt(np.sum(np.square(X[i] - X[j]), axis=-1).max())

    if lazy:
        X = X[
            np.random.choice(
                a=len(X),
                size=np.minimum(len(X), MAX_SAMPLE_SIZE),
                replace=False,
            )
        ]
    hull = sp.spatial.ConvexHull(X[:, dim_mask])
    vertices = hull.points[hull.vertices]
    i, j = np.triu_indices(len(vertices), k=1)

    return float(np.sqrt(np.max(np.square(vertices[i] - vertices[j]).sum(axis=-1))))


def get_rotation_matrix_2d(a):
    G = np.array([[0, -1], [1, 0]])  # generator
    A = np.expand_dims(a, axis=(-1, -2))  # angles
    return sp.linalg.expm(A * G)


def get_rotation_matrix_3d(**rotations):
    """
    A list of tuples [(dim, angle), ...] where we successively rotate around dim by angle.
    """
    dims = {"x": 0, "y": 1, "z": 2}
    R = np.eye(3)
    for axis, angle in rotations.items():
        i, j = (index for dim, index in dims.items() if dim != axis)
        S = np.zeros((*np.shape(angle), 3, 3))
        S[..., i, j] = angle
        R = sp.linalg.expm(S - np.swapaxes(S, -2, -1)) @ R
    return R


def get_orthogonal_transform(signature, entries):
    """
    A list of tuples [(dim, angle), ...] where we successively rotate around dim by angle.
    """

    axes = np.where(signature)[0]
    n_dim = len(signature)
    n_axes = sum(signature)

    if n_axes * (n_axes - 1) / 2 != len(entries):
        raise ValueError(
            f"Bad shape for entries (for signature {signature} we expect len(entries) = {int(n_axes * (n_axes - 1) / 2)}.",
        )

    i, j = np.triu_indices(n=n_axes, k=1)
    S = np.zeros((n_dim, n_dim))
    S[axes[i], axes[j]] = entries
    return sp.linalg.expm(S - S.T)


def compute_aligning_transform(points, signature, axes=None):
    """
    Find a transform for some (..., n_dim) array of points so that the volume over all but the first axis is minimized.
    """
    *input_shape, n_dim = points.shape
    axes = axes or tuple(range(1, n_dim))

    def loss(entries, *args):
        tp = args[0] @ get_orthogonal_transform(signature=signature, entries=entries)
        if n_dim > 2:
            ch = sp.spatial.ConvexHull(tp[..., 1:])
            return np.log(ch.volume)
        else:
            return np.log(np.ptp(tp[..., 1:]))

    n_axes = sum(signature)
    res = sp.optimize.minimize(
        loss,
        x0=np.zeros(int(n_axes * (n_axes - 1) / 2)),
        args=points.reshape(-1, n_dim),
        tol=1e-3,
        method="SLSQP",
    )

    if not res.success:
        raise RuntimeError("Could not find optimal rotation.")

    return get_orthogonal_transform(signature=signature, entries=res.x)



from __future__ import annotations

import numpy as np
import pytest

import maria
from maria import all_bands, all_regions
from maria.calibration import Calibration
from maria.constants import T_CMB
from maria.spectrum import AtmosphericSpectrum


def test_brightness_temperature_to_spectral_flux_density_per_pixel():
    square_arcminute = np.radians(1 / 60) ** 2

    assert np.isclose(
        Calibration("K_RJ -> Jy/pixel", nu=90e9, pixel_area=square_arcminute)(1e0),
        21.0576123,
    )


# def test_involution():

#     sim = Simulation(
#         instrument="MUSTANG-2",
#         plan="five_second_stare",
#         site="green_bank",
#         noise=True,
#     )

#     tod = sim.run()

#     assert np.isclose(
#         tod.signal.compute(), tod.to("K_RJ").to("pW").signal.compute()
#     ).all()

n_tests = 8
test_regions = np.random.choice(a=all_regions, size=n_tests)
test_bands = np.random.choice(a=all_bands, size=n_tests)


@pytest.mark.parametrize(
    "region,band_name",
    zip(test_regions, test_bands),
)
def test_cmb_atmosphere_reversability(region, band_name):
    eps = 1e-4
    kwargs_shape = (5, 7, 9)  # a random shape to confuse things

    band = maria.get_band(band_name)
    polarized = np.random.choice([True, False])
    spectrum = AtmosphericSpectrum(region=region)

    calibration_kwargs = {
        "spectrum": spectrum,
        "zenith_pwv": np.random.uniform(size=kwargs_shape, low=0, high=50),
        "base_temperature": np.random.uniform(
            size=kwargs_shape,
            low=spectrum.side_base_temperature.min(),
            high=spectrum.side_base_temperature.max(),
        ),
        "elevation": np.radians(np.random.uniform(size=kwargs_shape, low=10, high=90)),
        "polarized": polarized,
    }

    # calibration_kwargs = {}

    P_lo = Calibration("K_b -> fW", band=band, **calibration_kwargs)(T_CMB - eps / 2)
    P_hi = Calibration("K_b -> fW", band=band, **calibration_kwargs)(T_CMB + eps / 2)
    T = Calibration("fW -> K_CMB", band=band, **calibration_kwargs)(P_hi - P_lo)

    assert np.allclose(T, eps, rtol=1e-6)
